#ifndef _CHRONOPTICS_TOF_GIGE_INTERFACE_H_
#define _CHRONOPTICS_TOF_GIGE_INTERFACE_H_

#include <chronoptics/tof/camera_config.h>
#include <chronoptics/tof/calibration.h>
#include <chronoptics/tof/data.h>

#ifdef __cplusplus
extern "C" {
#endif

/** The possible depth streams that can be streamed from the camera
 */
enum tof_depth_stream_type {
  TOF_DEPTH_STREAM_TYPE_RAW = 0, /** Stream raw data */
  TOF_DEPTH_STREAM_TYPE_RADIAL_INTENSITY = 1, /** Stream both radial and intensity */
  TOF_DEPTH_STREAM_TYPE_RADIAL = 2, /** Stream radial */
  TOF_DEPTH_STREAM_TYPE_INTENSITY = 3, /** Stream intensity */
};

/** A gige discovery message containing camera information
 */
typedef struct tof_discovery_message* tof_discovery_message_t;

/** Destruct tof_discovery_message */
TOF_EXPORT void tof_discovery_message_delete(tof_discovery_message_t ptr);

/** Get the camera's ip address
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Ip address
 */
TOF_EXPORT const char* tof_discovery_message_ip(const tof_discovery_message_t ptr, tof_error_t *error);

/** Get the camera's serial
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Serial
 */
TOF_EXPORT const char* tof_discovery_message_serial(const tof_discovery_message_t ptr, tof_error_t *error);

/** A persistent ip data structure
 */
typedef struct tof_persistent_ip* tof_persistent_ip_t;

/** Destruct tof_persistent_ip */
TOF_EXPORT void tof_persistent_ip_delete(tof_persistent_ip_t ptr);

/** Create persistent ip
 * @param ptr Pointer to class
 * @param ip IP address
 * @param netmask Netmask
 * @param gateway Gateway
 * @param enabled Enable or disable persistent ip
 * @param error Pointer to error
 * @return New object
 */
TOF_EXPORT tof_persistent_ip_t tof_persistent_ip_new(const char* ip, const char* netmask, const char* gateway, bool enabled, tof_error_t *error);

/** Get the IP address
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return the IP address
 */
TOF_EXPORT const char* tof_persistent_ip_get_ip(const tof_persistent_ip_t ptr, tof_error_t *error);

/** Set the IP address
 * @param ptr Pointer to class
 * @param ip the IP address
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_ip(tof_persistent_ip_t ptr, const char* ip, tof_error_t *error);

/** Get the netmask
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The netmask
 */
TOF_EXPORT const char* tof_persistent_ip_get_netmask(const tof_persistent_ip_t ptr, tof_error_t *error);

/** Set the netmask
 * @param ptr Pointer to class
 * @param netmask The netmask
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_netmask(tof_persistent_ip_t ptr, const char* netmask, tof_error_t *error);

/** Get the gateway
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The gateway
 */
TOF_EXPORT const char* tof_persistent_ip_get_gateway(const tof_persistent_ip_t ptr, tof_error_t *error);

/** Set the gateway
 * @param ptr Pointer to class
 * @param gateway The gateway
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_gateway(tof_persistent_ip_t ptr, const char* gateway, tof_error_t *error);

/** Get if persistent ip is enabled
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Is enabled
 */
TOF_EXPORT bool tof_persistent_ip_get_enabled(const tof_persistent_ip_t ptr, tof_error_t *error);

/** Set if persistent ip is enabled
 * @param ptr Pointer to class
 * @param enabled Is enabled
 * @param error Pointer to error
 */
TOF_EXPORT void tof_persistent_ip_set_enabled(tof_persistent_ip_t ptr, bool enabled, tof_error_t *error);

/** Connect to a camera directly using the gige interface
 */
typedef struct tof_gige_interface* tof_gige_interface_t;

/** Destruct tof_gige_interface */
TOF_EXPORT void tof_gige_interface_delete(tof_gige_interface_t ptr);

/** Create an instance of the gige interface
 * @param ptr Pointer to class
 * @param port The port to use for GVCP communication, if 0 is specified it will
 * use an OS assigned port
 * @param error Pointer to error
 * @return New object
 */
TOF_EXPORT tof_gige_interface_t tof_gige_interface_new(uint16_t port, tof_error_t *error);

/** Detect all the cameras on the network
 * @param ptr Pointer to class
 * @param messages A list of discovery messages
 * @param capacity The amount of data the above pointer can hold
 * @param error Pointer to error
 * @return Number of pointers filled or complete array size
 */
TOF_EXPORT size_t tof_gige_interface_discover(tof_gige_interface_t ptr, tof_discovery_message_t* messages, size_t capacity, tof_error_t *error);

/** Returns a message of the first camera that responded
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return A discovery message
 */
TOF_EXPORT tof_discovery_message_t tof_gige_interface_discover_one(tof_gige_interface_t ptr, tof_error_t *error);

/** Detect a specific camera on the network, will throw an error if it can't
 * find the camera
 * @param ptr Pointer to class
 * @param serial Serial number of camera to find
 * @param error Pointer to error
 * @return A discovery message
 */
TOF_EXPORT tof_discovery_message_t tof_gige_interface_find(tof_gige_interface_t ptr, const char* serial, tof_error_t *error);

/** Connect to the camera and gain exclusive access
 * @param ptr Pointer to class
 * @param message Discovery message of the camera to connect to
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_connect(tof_gige_interface_t ptr, const tof_discovery_message_t message, tof_error_t *error);

/** Disconnect from the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_disconnect(tof_gige_interface_t ptr, tof_error_t *error);

/** Check whether a connection is still maintained with the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Is connected
 */
TOF_EXPORT bool tof_gige_interface_is_connected(const tof_gige_interface_t ptr, tof_error_t *error);

/** Get stream packet size
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The packet size
 */
TOF_EXPORT uint16_t tof_gige_interface_get_stream_packet_size(const tof_gige_interface_t ptr, tof_error_t *error);

/** Set stream packet size
 * @param ptr Pointer to class
 * @param packet_size The packet size
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_set_stream_packet_size(tof_gige_interface_t ptr, uint16_t packet_size, tof_error_t *error);

/** Test whether using the set packet size works
 * @param ptr Pointer to class
 * @param port Port to send test packets to
 * @param error Pointer to error
 * @return Packet size works
 */
TOF_EXPORT bool tof_gige_interface_test_stream_packet_size(tof_gige_interface_t ptr, uint16_t port, tof_error_t *error);

/** Get delay between network packets
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return The delay, depending on the camera software it is in nanoseconds or
 * in cpu instruction
 */
TOF_EXPORT uint32_t tof_gige_interface_get_delay(const tof_gige_interface_t ptr, tof_error_t *error);

/** Set delay between network packets
 * @param ptr Pointer to class
 * @param delay The delay, depending on the camera software it is in nanoseconds
 * or in cpu instruction
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_set_delay(tof_gige_interface_t ptr, uint32_t delay, tof_error_t *error);

/** Download camera configuration from the camera
 * @param ptr Pointer to class
 * @param port Port to download configuration on
 * @param error Pointer to error
 * @return Camera configuration
 */
TOF_EXPORT tof_camera_config_t tof_gige_interface_download_configuration(tof_gige_interface_t ptr, uint16_t port, tof_error_t *error);

/** Download calibration from the camera
 * @param ptr Pointer to class
 * @param port Port to download configuration on
 * @param error Pointer to error
 * @return Calibration
 */
TOF_EXPORT tof_calibration_t tof_gige_interface_download_calibration(tof_gige_interface_t ptr, uint16_t port, tof_error_t *error);

/** Upload camera configuration to the camera
 * @param ptr Pointer to class
 * @param config Camera configuration
 * @param port Port to download configuration on
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_upload_configuration(tof_gige_interface_t ptr, const tof_camera_config_t config, uint16_t port, tof_error_t *error);

/** Check whether this camera is capable of streaming the specified type
 * @param ptr Pointer to class
 * @param stream Depth stream type
 * @param error Pointer to error
 * @return Is capable
 */
TOF_EXPORT bool tof_gige_interface_depth_stream_capable(const tof_gige_interface_t ptr, enum tof_depth_stream_type stream, tof_error_t *error);

/** Start the depth stream
 * @param ptr Pointer to class
 * @param stream Depth stream type
 * @param port Port to receive depth data on
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_start_depth_stream(tof_gige_interface_t ptr, enum tof_depth_stream_type stream, uint16_t port, tof_error_t *error);

/** Stop the depth stream
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_stop_depth_stream(tof_gige_interface_t ptr, tof_error_t *error);

/** Returns true if a depth frame is available
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Depth frame available
 */
TOF_EXPORT bool tof_gige_interface_has_depth_data(const tof_gige_interface_t ptr, tof_error_t *error);

/** Get a depth frame
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Depth frame
 */
TOF_EXPORT tof_data_t tof_gige_interface_get_depth_data(tof_gige_interface_t ptr, tof_error_t *error);

/** Check whether camera supports an image stream
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Camera supports an image stream
 */
TOF_EXPORT bool tof_gige_interface_image_stream_capable(const tof_gige_interface_t ptr, tof_error_t *error);

/** Start image stream
 * @param ptr Pointer to class
 * @param port Port to receive image data on
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_start_image_stream(tof_gige_interface_t ptr, uint16_t port, tof_error_t *error);

/** Stop image stream
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_stop_image_stream(tof_gige_interface_t ptr, tof_error_t *error);

/** Check whether image frame is available
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Image frame available
 */
TOF_EXPORT bool tof_gige_interface_has_image_data(const tof_gige_interface_t ptr, tof_error_t *error);

/** Get image frame
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Image data
 */
TOF_EXPORT tof_data_t tof_gige_interface_get_image_data(tof_gige_interface_t ptr, tof_error_t *error);

/** Get the image frame into the supplied pointer
 * @param ptr Pointer to class
 * @param pointer The pointer to the data to write into
 * @param capacity The amount of data the pointer can hold
 * @param callback Callback that will be called when pointer is no longer in
 * use
 * @param callback_user_data User data that will be passed back when the function pointer is called
 * @param error Pointer to error
 * @return Image data
 */
TOF_EXPORT tof_data_t tof_gige_interface_get_image_data_into_pointer(tof_gige_interface_t ptr, uint8_t* pointer, size_t capacity, tof_user_pointer_destructed_t callback, void* callback_user_data, tof_error_t *error);

/** Get whether the camera is persistent ip capable
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Persistent IP capable
 */
TOF_EXPORT bool tof_gige_interface_persistent_ip_capable(const tof_gige_interface_t ptr, tof_error_t *error);

/** Get the current persistent ip settings
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Persistent IP
 */
TOF_EXPORT tof_persistent_ip_t tof_gige_interface_get_persistent_ip(const tof_gige_interface_t ptr, tof_error_t *error);

/** Set the persistent ip settings. The persistent ip settings will be active on
 * next reboot.
 * @param ptr Pointer to class
 * @param persistent_ip The persistent ip
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_set_persistent_ip(tof_gige_interface_t ptr, const tof_persistent_ip_t persistent_ip, tof_error_t *error);

/** Get the capacity for amount of user pointers that can be registered
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return User pointer capacity
 */
TOF_EXPORT size_t tof_gige_interface_get_user_pointer_capacity(const tof_gige_interface_t ptr, tof_error_t *error);

/** Set the capacity for amount of user pointers that can be registered
 * @param ptr Pointer to class
 * @param capacity User pointer capacity
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_set_user_pointer_capacity(tof_gige_interface_t ptr, size_t capacity, tof_error_t *error);

/** Add a pointer that will be filled with depth data. This method allows you to
 * fill data straight into your own data structure without any additional
 * copying.
 * @param ptr Pointer to class
 * @param pointer The pointer to the data to write into
 * @param capacity The amount of data the pointer can hold
 * @param callback Callback that will be called when pointer is no longer in
 * use
 * @param callback_user_data User data that will be passed back when the function pointer is called
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_add_depth_user_pointer(tof_gige_interface_t ptr, uint8_t* pointer, size_t capacity, tof_user_pointer_destructed_t callback, void* callback_user_data, tof_error_t *error);

/** Get the tof library version running on the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Version
 */
TOF_EXPORT const char* tof_gige_interface_version(tof_gige_interface_t ptr, tof_error_t *error);

/** Check whether the camera is capable of software trigger
 * @param ptr Pointer to class
 * @param error Pointer to error
 * @return Software trigger capable
 */
TOF_EXPORT bool tof_gige_interface_software_trigger_capable(tof_gige_interface_t ptr, tof_error_t *error);

/** Software trigger the camera
 * @param ptr Pointer to class
 * @param error Pointer to error
 */
TOF_EXPORT void tof_gige_interface_software_trigger(tof_gige_interface_t ptr, tof_error_t *error);

#ifdef __cplusplus
}
#endif

#endif
