#include <algorithm>
#include <chronoptics/tof/kea_camera.hpp>
#include <iostream>

namespace tof = chronoptics::tof;

struct XYZRGB {
  float x;
  float y;
  float z;
  uint8_t r;
  uint8_t g;
  uint8_t b;
};

template <typename T>
const T* get_pointer(const std::vector<tof::Data>& frames,
                     tof::FrameType type) {
  auto it = std::find_if(
      frames.begin(), frames.end(),
      [type](const tof::Data& frame) { return frame.frame_type() == type; });

  if (it == frames.end())
    throw std::runtime_error("Couldn't find type in supplied frames");

  const tof::Data& data = *it;
  return data.begin<T>();
}

// This function is a simple implementation and will only create a point cloud
// of the first x, y, z and bgr in the frames
std::vector<XYZRGB> create_point_cloud(const std::vector<tof::Data>& frames) {
  // Get pointers to the various data types
  const float* x = get_pointer<float>(frames, tof::FrameType::X);
  const float* y = get_pointer<float>(frames, tof::FrameType::Y);
  const float* z = get_pointer<float>(frames, tof::FrameType::Z);
  const uint8_t* bgr =
      get_pointer<uint8_t>(frames, tof::FrameType::BGR_PROJECTED);

  // Calculate size of the image
  auto& frame = frames.front();
  const size_t size = frame.rows() * frame.cols();

  // Preallocate required size
  std::vector<XYZRGB> point_cloud;
  point_cloud.reserve(size);

  // Loop over each pixel and create the XYZRGB struct. Note that the bgr is
  // stored the opposite way of rgb
  for (size_t i = 0; i < size; i++) {
    point_cloud.push_back(
        XYZRGB{x[i], y[i], z[i], bgr[i * 3 + 2], bgr[i * 3 + 1], bgr[i * 3]});
  }

  return point_cloud;
}

void use_point_cloud(std::vector<XYZRGB>&) {
  // Do something
}

int main(int argc, char const** argv) {
  if (argc != 2) {
    std::cout << "Expected " << argv[0] << " <serial>\n";
    return -1;
  }

  tof::KeaCamera cam(tof::ProcessingConfig{}, argv[1]);

  auto stream_list = cam.get_stream_list();

  const auto has_type = [&](tof::FrameType type) {
    auto it = std::find_if(
        stream_list.begin(), stream_list.end(),
        [type](tof::Stream& stream) { return stream.frame_type() == type; });
    return it != stream_list.end();
  };

  if (!has_type(tof::FrameType::X)) {
    std::cout << "Camera has incorrect configuration\n";
    return -1;
  }

  if (!has_type(tof::FrameType::BGR_PROJECTED)) {
    std::cout << "Camera doesn't have rgb support\n";
    return -1;
  }

  tof::select_streams(cam, {tof::FrameType::X, tof::FrameType::Y,
                            tof::FrameType::Z, tof::FrameType::BGR_PROJECTED});

  cam.start();

  // Showing how you would normally loop
  for (size_t i = 0; i < 1; i++) {
    auto frames = cam.get_frames();
    auto point_cloud = create_point_cloud(frames);
    use_point_cloud(point_cloud);
  }

  cam.stop();

  return 0;
}
